
; ****************************************************************************
;
; BBM OS KERNEL
;
;   Startup Assembly Routines
;
; Copyright (C) Sukru TIKVES 2006
;
;   Based on parts by GRUB, FritzOS and others, see LICENCE for information
;
; ****************************************************************************

; ****************************************************************************
;  GLOBAL SYMBOLS
; ****************************************************************************

GLOBAL KernelStart			; Start address for use by the linker

GLOBAL UpdateGDT		; To change GDTR from C/C++ code

GLOBAL GetEBP				; Returns EBP register
GLOBAL StackLimit			; Upper limit of stack
					; NOTE: stack grows in reverse
					;   direction, this is the initial
					;   value.

; ****************************************************************************
;  EXTERNAL SYMBOLS
; ****************************************************************************

EXTERN main						; Startup address of C/C++ main code
EXTERN keyboard_handler	; Keyboard handler
EXTERN timer_handler		; Timer Handler
EXTERN remap_pics			; Remap Pic

; ****************************************************************************
;  MULTI BOOT HEADER INFORMATION
; ****************************************************************************

MODULE_ALIGN equ  01h			; Align modules on page boundaries
MEMINFO      equ  02h			; Provide memory map
FLAGS        equ  MODULE_ALIGN | MEMINFO; Our flag settings
MAGIC        equ  0x1BADB002		; 'Magic number' for recognition
CHECKSUM     equ -(MAGIC + FLAGS)	; Checksum (required by bootleader)

section .text				; CODE SECTION
align 4

MultiBootHeader:
	dd	MAGIC
	dd	FLAGS
	dd	CHECKSUM

KernelStart:
	lgdt	[ GDTR ]		; Inform CPU about GTD table
	
	call remap_pics		; Remap IRQ's to prevent collusion with reserved ones 
	call set_ISR			; Set all IDT entry offsets
	lidt [ IDTR ]			; Inform CPU about IDT table
	
	
	
	sti 					; It is now safe to open interrupts

	mov	ax, DATA_SEGMENT	; Initialize segment registers
	mov	ds, ax						;   by setting to new data selector
	mov	es, ax
	mov	fs, ax
	mov	gs, ax

	mov	ss, ax				; Initialize kernel stack
	mov	eax, StackLimit	; End of stack (moves backwards)
	mov	esp, eax

	xor	ebp, ebp

	jmp	CODE_SEGMENT:Continue	; Load new selector in CS

Continue:
	push	ebx			; Save address of multiboot info
	call	main

Halt:
	jmp	Halt			; Main should not return

UpdateGDT
	mov	eax, [esp+4]
	mov	[GDTR+2], eax
	mov	ax, [esp+8]
	mov	[GDTR], ax
	lgdt	[GDTR]
	ret

GetEBP
	mov	eax, ebp
	ret
	
; Set all ISRs
set_ISR	
	pusha
	push gs
 	push fs
 	push es
 	push ds
	
	; Set timer ISR
	mov eax, timer_int			
	mov [IDT + timer_irq],ax	
	shr eax,16
	mov [IDT+ timer_irq + 6],ax

	; Set keyboard ISR
	mov eax, keyboard_int	
	mov [IDT + keyboard_irq],ax	
	shr eax,16
	mov [IDT + keyboard_irq + 6],ax
	
	pop ds
 	pop es
 	pop fs
 	pop gs
	popa
	ret

; Timer intterrupt
timer_int:					
	pushad				
	call timer_handler 	; Call handler from kernel
	mov al,20h			
	out 20h,al
	popad
	iret

; Keyboard interrupt
keyboard_int:						
	pushad				
	call keyboard_handler ; Call handler from kernel
	mov al,20h			
	out 20h,al
	;mov al,20h	; for extra safety	
	;out 20h,al		;  "     "        "    
	popad
	iret


; ****************************************************************************
;  GDT & IDT TABLES
; ****************************************************************************
section .data


; **** GDTR ****

GDTR
	dw	01fh
	dd	GDT

GDT
NULL_SEGMENT	equ 0
	dd 	0
	dd	0
CODE_SEGMENT	equ 8
	dw 0ffffh
	dw 0
	db 0
	db 09ah
	db 0cfh
	db 0h
DATA_SEGMENT	equ 010h
	dw 0ffffh
	dw 0h
	db 0
	db 092h
	db 0cfh
	db 0


; **** IDTR ****
; Table is partly formed from the tutorial below...
; http://www.osdever.net/tutorials/interrupts.3.php?the_id=41

IDTR 
	dw IDT_end - IDT - 1
	dd IDT
	
IDT
%rep 20h	; First 32 ISRs is for CPU exceptions...
  	dw 00h   
  	dw 08h 
  	dw 0E00h 
  	dw 08h 
%endrep
timer_irq equ $ - IDT	; Timer interrupt
	dw 00h 		; Offset 15-0(word)
	dw 08h 		; Selector31-16(word)
	dw 8E00h 	; Present(1 bit)	DPL(2 bits)	01110(5 bits)	000(3 bits)	Not Used
	dw 00h 		; Offset 31-16(word)
	
keyboard_irq equ $ - IDT	; Keyboard interrupt
	dw 00h		; Offset 15-0(word)
	dw 08h		; Selector31-16(word)
	dw 8E00h	; Present(1 bit)	DPL(2 bits)	01110(5 bits)	000(3 bits)	Not Used
	dw 00h		; Offset 31-16(word)
	
%rep 0DDh	; Remaining ones
  	dw 00h
  	dw 08h
  	dw 0E00h
  	dw 08h
%endrep
IDT_end


; ****************************************************************************
;  KERNEL STACK
; ****************************************************************************

STACK_SIZE equ 0x4000			; 16K stack size

section .bss

stack
	resb STACK_SIZE
StackLimit:
